# Copyright Alan (AJ) Pryor, Jr. 2017
# Transcribed from MATLAB code by Colin Ophus
# Prismatic is distributed under the GNU General Public License (GPL)
# If you use Prismatic, we kindly ask that you cite the following papers:

# 1. Ophus, C.: A fast image simulation algorithm for scanning
#    transmission electron microscopy. Advanced Structural and
#    Chemical Imaging 3(1), 13 (2017)

# 2. Pryor, Jr., A., Ophus, C., and Miao, J.: A Streaming Multi-GPU
#    Implementation of Image Simulation Algorithms for Scanning
#	 Transmission Electron Microscopy. arXiv:1706.08563 (2017)

from . import core
from . import fileio
from pyprismatic.params import Metadata

def demo():
	import os 
	with open('temp.XYZ', 'w') as fid:
		fid.write("one unit cell of 100 silicon\n\
  5.43    5.43    5.43\n\
14  0.0000  0.0000  0.0000  1.0  0.076\n\
14  2.7150  2.7150  0.0000  1.0  0.076\n\
14  1.3575  4.0725  1.3575  1.0  0.076\n\
14  4.0725  1.3575  1.3575  1.0  0.076\n\
14  2.7150  0.0000  2.7150  1.0  0.076\n\
14  0.0000  2.7150  2.7150  1.0  0.076\n\
14  1.3575  1.3575  4.0725  1.0  0.076\n\
14  4.0725  4.0725  4.0725  1.0  0.076\n\
-1")
	meta = Metadata(filenameAtoms='temp.XYZ',
				    filenameOutput='output.mrc');
	meta.algorithm = 'multislice'
	meta.go()
	import numpy as np
	from pyprismatic.fileio import readMRC
	import matplotlib.pyplot as plt
	result = readMRC("output.mrc")
	plt.figure()
	plt.imshow(np.squeeze(np.sum(result,axis=2)))
	plt.show()
	os.remove("temp.XYZ")


# Copyright Alan (AJ) Pryor, Jr. 2017
# Transcribed from MATLAB code by Colin Ophus
# Prismatic is distributed under the GNU General Public License (GPL)
# If you use Prismatic, we kindly ask that you cite the following papers:

# 1. Ophus, C.: A fast image simulation algorithm for scanning
#    transmission electron microscopy. Advanced Structural and
#    Chemical Imaging 3(1), 13 (2017)

# 2. Pryor, Jr., A., Ophus, C., and Miao, J.: A Streaming Multi-GPU
#    Implementation of Image Simulation Algorithms for Scanning
#	 Transmission Electron Microscopy. arXiv:1706.08563 (2017)

def readMRC(filename, dtype=float, order="C"):
    """
    * readMRC *

    Read in a volume in .mrc file format. See http://bio3d.colorado.edu/imod/doc/mrc_format.txt

    :param filename: Filename of .mrc
    :return: NumPy array containing the .mrc data

    Author: Alan (AJ) Pryor, Jr.
    Jianwei (John) Miao Coherent Imaging Group
    University of California, Los Angeles
    Copyright 2015-2016. All rights reserved.

    """
    import numpy as np
    import struct
    headerIntNumber = 56
    sizeof_int = 4
    headerCharNumber = 800
    sizeof_char = 1
    with open(filename,'rb') as fid:
        int_header = struct.unpack('=' + 'i'*headerIntNumber, fid.read(headerIntNumber * sizeof_int))
        char_header = struct.unpack('=' + 'c'*headerCharNumber, fid.read(headerCharNumber * sizeof_char))
        dimz, dimy, dimx, data_flag= int_header[:4]
        if (data_flag == 0):
            datatype='u1'
        elif (data_flag ==1):
            datatype='i1'
        elif (data_flag ==2):
            datatype='f4'
        elif (data_flag ==3):
            datatype='c'
        elif (data_flag ==4):
            datatype='f4'
        elif (data_flag ==6):
            datatype='u2'
        else:
            raise ValueError("No supported datatype found!\n")

        return np.fromfile(file=fid, dtype=datatype,count=dimx*dimy*dimz).reshape((dimx,dimy,dimz),order=order).astype(dtype)



# Copyright Alan (AJ) Pryor, Jr. 2017
# Transcribed from MATLAB code by Colin Ophus
# Prismatic is distributed under the GNU General Public License (GPL)
# If you use Prismatic, we kindly ask that you cite the following papers:

# 1. Ophus, C.: A fast image simulation algorithm for scanning
#    transmission electron microscopy. Advanced Structural and
#    Chemical Imaging 3(1), 13 (2017)

# 2. Pryor, Jr., A., Ophus, C., and Miao, J.: A Streaming Multi-GPU
#    Implementation of Image Simulation Algorithms for Scanning
#	 Transmission Electron Microscopy. arXiv:1706.08563 (2017)

import pyprismatic.core
class Metadata(object):
	"""
	"interpolationFactorX" : PRISM interpolation factor in x-direction
	"interpolationFactorY" : PRISM interpolation factor in y-direction
	"filenameAtoms" : filename containing input atom information in XYZ format (see http://prism-em.com/about/ for more details)
	"filenameOutput" : filename in which to save the 3D output. Also serves as base filename for 2D and 4D outputs if used
	"realspacePixelSizeX" : size of pixel size in X for probe/potential arrays
	"realspacePixelSizeY" : size of pixel size in Y for probe/potential arrays
	"potBound" : limiting radius within which to compute projected potentials from the center of each atom (in Angstroms)
	"numFP" : number of frozen phonon configurations to average over
	"sliceThickness" : thickness of potential slices (in Angstroms)
	"cellDimX" : unit cell dimension X (in Angstroms)
	"cellDimY" : unit cell dimension Y (in Angstroms)
	"cellDimZ" : unit cell dimension Z (in Angstroms)
	"tileX" : number of unit cells to tile in X direction
	"tileY" : number of unit cells to tile in Y direction
	"tileZ" : number of unit cells to tile in Z direction
	"E0" : electron beam energy (in eV)
	"alphaBeamMax" : the maximum probe angle to consider (in mrad)
	"numGPUs" : number of GPUs to use. A runtime check is performed to check how many are actually available, and the minimum of these two numbers is used. 
	"numStreamsPerGPU" : number of CUDA streams to use per GPU
	"numThreads" : number of CPU worker threads to use
	"batchSizeTargetCPU" : desired batch size for CPU FFTs.
	"batchSizeTargetGPU" : desired batch size for GPU FFTs.
	"earlyCPUStopCount" : the WorkDispatcher will cease providing work to CPU workers earlyCPUStopCount jobs from the end. This is to prevent the program waiting for slower CPU workers to complete.
    "probeStepX" : step size of the probe in X direction (in Angstroms)
	"probeStepY" : step size of the probe in Y direction (in Angstroms)
	"probeDefocus" : probe defocus (in Angstroms)
	"C3" : microscope C3 (in Angstroms)
	"C5" : microscope C5 (in Angstroms)
	"probeSemiangle" : probe convergence semi-angle (in mrad)
	"detectorAngleStep" : angular step size for detector integration bins (in mrad)
	"probeXtilt" : (in Angstroms)
	"probeYtilt" : (in Angstroms)
	"scanWindowXMin" : lower X size of the window to scan the probe (in fractional coordinates)
	"scanWindowXMax" : upper X size of the window to scan the probe (in fractional coordinates)
	"scanWindowYMin" : lower Y size of the window to scan the probe (in fractional coordinates)
	"scanWindowYMax" : upper Y size of the window to scan the probe (in fractional coordinates)
	"randomSeed" : number to use for random seeding of thermal effects
	"algorithm" : simulation algorithm to use, "prism" or "multislice"
	"includeThermalEffects" : true/false to apply random thermal displacements (Debye-Waller effect)
	"alsoDoCPUWork" : true/false
	"save2DOutput" : save the 2D STEM image integrated between integrationAngleMin and integrationAngleMax
	"save3DOutput" : true/false Also save the 3D output at the detector for each probe (3D output mode)
	"save4DOutput" : true/false Also save the 4D output at the detector for each probe (4D output mode)
	"integrationAngleMin" : (in mrad)
	"integrationAngleMax" : (in mrad)
	"transferMode : memory model to use, either "streaming", "singlexfer", or "auto"
	"""

	fields=[
	"interpolationFactorX",
	"interpolationFactorY",
	"filenameAtoms",
	"filenameOutput",
	"realspacePixelSizeX",
	"realspacePixelSizeY",
	"potBound",
	"numFP",
	"sliceThickness",
	"cellDimX",
	"cellDimY",
	"cellDimZ",
	"tileX",
	"tileY",
	"tileZ",
	"E0",
	"alphaBeamMax",
	"numGPUs",
	"numStreamsPerGPU",
	"numThreads",
	"batchSizeTargetCPU",
	"batchSizeTargetGPU",
	"earlyCPUStopCount",
    "probeStepX",
	"probeStepY",
	"probeDefocus",
	"C3",
	"C5",
	"probeSemiangle",
	"detectorAngleStep",
	"probeXtilt",
	"probeYtilt",
	"scanWindowXMin",
	"scanWindowXMax",
	"scanWindowYMin",
	"scanWindowYMax",
	"randomSeed",
	"algorithm",
	"includeThermalEffects",
	"alsoDoCPUWork",
	"save2DOutput",
	"save3DOutput",
	"save4DOutput",
	"integrationAngleMin",
	"integrationAngleMax",
	"transferMode"]
	def __init__(self, *args, **kwargs):
		"""
		Fields within Metadata objects can be set either manually or at construction time. For example:

		meta = Metadata(interpolationFactorX=8, filenameOutput="test.mrc") 

		would initialize a Metadata object with all parameters set to defaults except for interpolationFactorX and filenameOutput, which take on the values 8 and "test.mrc"

		"""
		import numpy as np
		self.interpolationFactorX 	  = 4
		self.interpolationFactorY 	  = 4
		self.filenameAtoms		      = ""
		self.filenameOutput	          = "output.mrc"
		self.realspacePixelSizeX      = 0.1
		self.realspacePixelSizeY      = 0.1
		self.potBound				  = 1.0
		self.numFP 				      = 1
		self.sliceThickness		      = 2.0
		self.cellDimX                 = 20.0
		self.cellDimY                 = 20.0
		self.cellDimZ                 = 20.0
		self.tileX					  = 3
		self.tileY					  = 3
		self.tileZ					  = 1
		self.E0						  = 80e3
		self.alphaBeamMax			  = 0.024
		self.numGPUs     			  = 4
		self.numStreamsPerGPU         = 3	
		self.numThreads  		      = 12
		self.batchSizeTargetCPU	      = 1
		self.batchSizeTargetGPU       = 2
		self.batchSizeCPU             = 1
		self.batchSizeGPU             = 1
		self.earlyCPUStopCount        = 100.0
		self.probeStepX		          = 0.25
		self.probeStepY			      = 0.25
		self.probeDefocus			  = 0.0
		self.C3						  = 0.0
		self.C5						  = 0.0
		self.probeSemiangle			  = 0.02
		self.detectorAngleStep	      = 0.001
		self.probeXtilt				  = 0.0
		self.probeYtilt				  = 0.0
		self.scanWindowXMin			  = 0.0
		self.scanWindowXMax			  = 1.0
		self.scanWindowYMin  		  = 0.0
		self.scanWindowYMax			  = 1.0
		self.randomSeed	     	      = np.random.randint(0,999999)
		self.algorithm				  = "prism"
		self.includeThermalEffects    = False
		self.alsoDoCPUWork		      = True
		self.save2DOutput			  = False
		self.save3DOutput		      = True
		self.save4DOutput			  = False
		self.integrationAngleMin      = 0
		self.integrationAngleMax      = .001
		self.transferMode		      = "auto"
		for k,v in kwargs.items():
			if k not in Metadata.fields:
				print("Invalid metaparameter \"{}\" provided".format(k))
			else:
				setattr(self, k, v)
	def toString(self):
			for field in Metadata.fields:
				print("{} = {}".format(field, getattr(self, field)))
	def go(self):
		self.algorithm    = self.algorithm.lower()
		self.transferMode = self.transferMode.lower()
		l = [getattr(self, field) for field in Metadata.fields]
		pyprismatic.core.go(*(l))



# Copyright Alan (AJ) Pryor, Jr. 2017
# Transcribed from MATLAB code by Colin Ophus
# Prismatic is distributed under the GNU General Public License (GPL)
# If you use Prismatic, we kindly ask that you cite the following papers:

# 1. Ophus, C.: A fast image simulation algorithm for scanning
#    transmission electron microscopy. Advanced Structural and
#    Chemical Imaging 3(1), 13 (2017)

# 2. Pryor, Jr., A., Ophus, C., and Miao, J.: A Streaming Multi-GPU
#    Implementation of Image Simulation Algorithms for Scanning
#    Transmission Electron Microscopy. arXiv:1706.08563 (2017)

from setuptools import setup, Extension
from setuptools.command.install import install
import sys
import os

prismatic_extra_definitions = []
prismatic_libs  		    = []

# In CPU-only mode, all Prismatic C++ source files will be compiled into the Python package. For GPU support, a shared library is
# built from the CUDA/C++ sources and the Python package links against it. The C++ files that are potentially compiled into
# the shared library (along with CUDA files) are the "extra" sources and in CPU-mode they are added to the Python package sources.
# The reason for doing it this way is for user convenience so that compiling/linking to an extra Prismatic library is not required until
# enabling for GPU, at which point it is already necessary for the other CUDA libraries.
prismatic_sources_base  = ['pyprismatic/core.cpp'
						  ]
prismatic_sources_extra = [
	'src/go.cpp',
	'src/atom.cpp',
	'src/configure.cpp',
	'src/Multislice_calcOutput.cpp',
	'src/Multislice_entry.cpp',#
	'src/parseInput.cpp',#
	'src/PRISM01_calcPotential.cpp',
	'src/PRISM02_calcSMatrix.cpp',#
	'src/PRISM03_calcOutput.cpp',#
	'src/PRISM_entry.cpp',
	'src/projectedPotential.cpp',#
	'src/utility.cpp',#
	'src/WorkDispatcher.cpp'
							]
prismatic_sources 		    = prismatic_sources_base
prismatic_include_dirs 		= ["./include"]
prismatic_library_dirs 		= []

if os.name == "nt": #check for Windows OS
	prismatic_libs 		   		= ["libfftw3f-3"]
else:
	prismatic_libs 		   		= ["fftw3f", "fftw3f_threads"]
prismatic_extra_compile_defs= ['-std=c++11']

class InstallCommand(install):
	user_options = install.user_options + [
		('enable-gpu', None, None),
	]

	def initialize_options(self):
		install.initialize_options(self)
		self.enable_gpu = False

	def finalize_options(self):
		install.finalize_options(self)

	def run(self):
		install.run(self)


if ("--enable-gpu" in sys.argv): # GPU-mode, add some macro definitions and link with the CUDA libraries
	prismatic_extra_definitions.extend([("PRISMATIC_ENABLE_GPU",1), ("BUILDING_CUPRISMATIC",1)])
	prismatic_libs.extend(["cuprismatic", "cufft", "cudart"])
else: # CPU-only mode, add the extra C++ sources to the python package
	prismatic_sources.extend(prismatic_sources_extra)

pyprimsatic_core = Extension('pyprismatic.core',
	sources=prismatic_sources,
	include_dirs=prismatic_include_dirs,
	extra_compile_args=prismatic_extra_compile_defs,
	define_macros=prismatic_extra_definitions,
	library_dirs=prismatic_library_dirs,
	libraries=prismatic_libs)

setup(name = 'PyPrismatic',
	author = 'Alan (AJ) Pryor, Jr.', 
	author_email='apryor6@gmail.com',
	version = '1.1.16',
	description="Python wrapper for Prismatic package for fast image simulation using the PRISM and multislice algorithms in Scanning Transmission Electron Microscopy (STEM)",
	ext_modules=[pyprimsatic_core],
	packages=['pyprismatic'],
    install_requires=['numpy>=1.13.0','matplotlib>=2.0.2'],
    cmdclass={'install': InstallCommand})



# This script calculates multiple images using a unit cell tiled varying numbers of times in Z. 
# This type of simulation is useful for determining the true thickness of a sample from experimental images

import pyprismatic as pr
import numpy as np
import matplotlib.pyplot as plt

base_output_name = "thickness_scan"
base_output_ext  = ".mrc"
atom_filename    = "../SI100.XYZ"
tileX=tileY=3
meta = pr.Metadata(filenameAtoms=atom_filename, tileX=tileX, tileY=tileY)
output_filenames=[]

# run simulations
for tileZ in range(1,17):
	meta.tileZ = tileZ
	meta.filenameOutput = base_output_name + str(tileZ) + base_output_ext
	output_filenames.extend([meta.filenameOutput])
	meta.go()


# print the results
f, ax = plt.subplots(4, 4)
ax = ax.ravel()
for i, filename in enumerate(output_filenames):
	stack = pr.fileio.readMRC(filename)
	img   = np.sum(stack[:, :, :10], axis=2)
	ax[i].imshow(img)
	ax[i].set_title("Thicknesss: {:.2f}$\AA$".format((i + 1) * 5.43))
	ax[i].set_yticklabels([])
	ax[i].set_xticklabels([])
plt.suptitle("Bright Field (0-10 mrad) PRISM images for SI100 of varying thicknesses")
plt.show()
